from dash_extensions.enrich import (
    dcc,
    html,
    DashBlueprint,
    LogTransform,
    NoOutputTransform,
    ServersideOutputTransform
)

import dash
import dash_mantine_components as dmc

from kmviz.ui import state
from kmviz.ui.id_factory import kmviz_factory as kf
from kmviz.ui.components.store import make_stores, make_stores_callbacks
from kmviz.ui.components.select import make_select, make_select_callbacks
from kmviz.ui.layouts.sidebar import make_sidebar_layout, make_sidebar_layout_callbacks
from kmviz.ui.layouts.tabs import make_tabs
import uuid

def make_dashboard():

    layout = html.Div([
        dcc.Store(kf.sid("session-id")),
        make_stores(),
        make_select(),
        dmc.Grid([
            dmc.Col(make_sidebar_layout(), id=kf.sid("sidebar-layout"), span="content"),
            dmc.Col(make_tabs(), id=kf.sid("tabs-layout"), span="auto")
        ])
    ])

    make_stores_callbacks()
    make_select_callbacks()
    make_sidebar_layout_callbacks()

    return layout
    #return dashboard

dash.register_page(__name__, path=state.kmstate.dashboard_path, name="dashboard", title="dashboard")

layout = make_dashboard()



from dash_extensions.enrich import html, dash_table, Input, Output, dcc, State, callback
import dash_mantine_components as dmc
from dash import no_update
import plotly.express as px
import plotly.graph_objects as go

from kmviz.ui.utils import make_select_data, prevent_update_on_none
from kmviz.ui.components.figure import make_plot_title, make_plot_title_callbacks
from kmviz.ui.components.select import kgsf
from kmviz.ui.components.store import ksf
from kmviz.ui.layouts.table import ktable

from kmviz.ui.id_factory import kmviz_factory as kf

kseq = kf.child("seq")

def blank_figure():
    fig = go.Figure(go.Scatter(x=[], y = []))
    fig.update_layout(template = "seaborn")
    fig.update_xaxes(showgrid = False, showticklabels = False, zeroline=False)
    fig.update_yaxes(showgrid = False, showticklabels = False, zeroline=False)

    return fig

def make_sequence_layout():
    res =  html.Div([
        dmc.Group([
            dmc.Select(
                id=kseq.sid("select"),
                label="Select ID"
            )
        ]),

        dcc.Graph(
            figure=blank_figure(),
            id=kseq.sid("figure"),
            responsive=True,
            style = { "margin-left": "auto", "margin-right": "auto", "height": "50vh", "width": "95%" }
        ),

        make_plot_title(kseq.child("title")),
    ])

    return res

def make_sequence_layout_callbacks():

    @callback(
        Input(kgsf("provider"), "value"),
        Input(kgsf("query"), "value"),
        State(ksf("query-results"), "data"),
        Output(kseq.sid("select"), "data"),
        Output(kseq.sid("select"), "value"),
        Output(kseq.sid("panel"), "disabled"),
        prevent_initial_callbacks=True,
    )
    def update_sample_select(provider, query, query_result):
        prevent_update_on_none(provider, query)

        if provider.startswith("__kmviz_df"):
            prevent_update_on_none(None)

        qr = query_result[query][provider]
        ids = list(qr.df["ID"])
        if not ids:
            return no_update, no_update, True
        return make_select_data(ids), ids[0], False

    @callback(
        Input(ktable.sid("grid"), "selectedRows"),
        Output(kseq.sid("select"), "value"),
        Output("tab-select", "value"),
        prevent_initial_callbacks=True
    )
    def on_selected(data):
        if not len(data):
            prevent_update_on_none(None)

        return data[0]["ID"], "sequence"

    @callback(
        Input(kseq.sid("select"), "value"),
        Input(kgsf("provider"), "value"),
        Input(kgsf("query"), "value"),
        State(ksf("query-results"), "data"),
        Output(kseq.sid("figure"), "figure"),
        prevent_initial_callbacks=True,
    )
    def update_sequence_graph(sample, provider, query, query_result):
        if provider.startswith("__kmviz_df"):
            prevent_update_on_none(None)

        prevent_update_on_none(sample, provider, query)

        qr = query_result[query][provider]

        f = {'ID': {'filterType': 'text', 'type': 'contains', 'filter': sample}}

        index = list(range(len(qr.query.seq)))

        if qr.response[sample].has_abs():
            fig = px.line(x=index, y=qr.response[sample].covyb, line_shape="linear")
        else:
            fig = px.line(x=index, y=qr.response[sample].covxb, line_shape="hv")

        fig.update_layout(
            xaxis = dict(
                tickmode = 'array',
                tickvals = index,
                ticktext = [x for x in qr.query.seq]
            )
        )
        fig.update_layout(
            xaxis=dict(rangeslider=dict(visible=True,range=[0, len(qr.query.seq)]),
                       range=[0, 50],
                       type="category")
        )

        return fig

    make_plot_title_callbacks(kseq.child("title"), kseq.sid("figure"))




from dash_extensions.enrich import html, dash_table, Input, Output, dcc, State, callback, ctx
import dash_mantine_components as dmc
from dash_iconify import DashIconify

import dash_ag_grid as dag
from dash import Patch
from kmviz.ui import state
import pandas as pd

from kmviz.ui.components.grid import make_ag_grid
from kmviz.ui.id_factory import kmviz_factory as kf
from kmviz.ui.utils import prevent_update_on_none
from kmviz.ui.components.select import kgsf
from kmviz.ui.components.store import ksf

from pandas.api.types import is_numeric_dtype, is_datetime64_dtype

ktable = kf.child("table")

def make_table_layout():
    res = html.Div([
        dmc.Space(h=5),
        make_ag_grid(ktable.sid("grid"), {}, {}, {}, style = {"height": "85vh"}),
        dmc.Space(h=5),
        dmc.Group([
            dmc.Button(
                "Export",
                id=ktable.sid("button"),
                disabled=True,
                leftIcon=DashIconify(icon="ph:export", width=20)
            ),
            dmc.Button(
                "Remove filters",
                id=ktable.sid("rmf"),
                disabled=True,
                leftIcon=DashIconify(icon="ph:trash", width=20)
            ),
            dmc.Button(
                "Filter NaN",
                id=ktable.sid("nan"),
                disabled=True,
                leftIcon=DashIconify(icon="mingcute:na-fill", width=20)
            )
        ])
    ])

    return res

def make_table_layout_callbacks():

    @callback(
        Input(kgsf("query"), "value"),
        Input(kf.sid("plot-only"), "data"),
        State(kgsf("provider"), "value"),
        State(ksf("query-results"), "data"),
        Output(ktable.sid("grid"), "rowData"),
        Output(ktable.sid("grid"), "columnDefs"),
        Output(ktable.sid("panel"), "disabled"),
        prevent_initial_call=True,
    )
    def update_table_grid(query, plot_only, provider, query_result):

        def column_filter(data):
            if is_numeric_dtype(data):
                return "agNumberColumnFilter"
            if is_datetime64_dtype(data):
                return "agDateColumnFilter"
            return "agTextColumnFilter"

        if not plot_only:
            df = query_result[query][provider].df
        else:
            df = plot_only["df"]

        fields = [
            {
                "field": x,
                "filterParams": {"maxNumConditions": 10000},
                "suppressMenu": True,
                #"filter": column_filter(query_result[query][provider].df[x])
                "filter": column_filter(df[x])
            }
            #for x in list(query_result[query][provider].df)
            for x in list(df)
        ]

        #return query_result[query][provider].df.to_dict("records"), fields, False
        return df.to_dict("records"), fields, False

    @callback(
        Input(ktable.sid("button"), "n_clicks"),
        Output(ktable.sid("grid"), "exportDataAsCsv"),
        prevent_initial_callbacks=True,
    )
    def download_table_as_csv(n_clicks):
        if n_clicks:
            return True
        prevent_update_on_none(None)

    @callback(
        Input(kgsf("query"), "value"),
        State(kgsf("provider"), "value"),
        Input(kf.sid("plot-only"), "data"),
        Output(ktable.sid("button"), "disabled"),
        Output(ktable.sid("rmf"), "disabled"),
        Output(ktable.sid("nan"), "disabled"),
        prevent_initial_callbacks=True,
    )
    def enable_table_grid_buttons(query, provider, po):
        prevent_update_on_none(query, provider)
        return False, False, False

    @callback(
        Input(ktable.sid("rmf"), "n_clicks"),
        Output(ktable.sid("grid"), "filterModel")
    )
    def remove_table_filters(n_clicks):
        if n_clicks:
            return {}
        prevent_update_on_none(None)

    @callback(
        Input(ktable.sid("nan"), "n_clicks"),
        State(ktable.sid("grid"), "columnDefs"),
        Output(ktable.sid("grid"), "filterModel", allow_duplicate=True),
    )
    def filter_nan(n_clicks, fields):
        if n_clicks:
            cols = [x["field"] for x in fields]
            p = Patch()
            for c in cols[1:]:
                p[c] = {'filterType': 'text', 'type': 'notBlank', 'filter': ''}
            return p






import dash_mantine_components as dmc
from dash_extensions.enrich import Input, Output, html, callback
from dash import no_update
from kmviz.ui.utils import prevent_update_on_empty, make_select_data, prevent_update_on_none, icons
from kmviz.ui.id_factory import kmviz_factory as kf
from kmviz.ui.components.store import ksfr
from kmviz.ui import state

kpsf = kf.child("select")
kgsf = kf.child("global-select")

from dash_iconify import DashIconify

def make_select_provider():
    res = html.Div([
        dmc.Divider(size="sm", color="gray", label="DATABASES", labelPosition="center"),
        dmc.MultiSelect(
            id=kpsf("provider"),
            label="Database(s)",
            data=make_select_data(state.kmstate.providers.list()),
            clearable=True,
            withAsterisk=True,
            icon=DashIconify(icon="fluent:library-16-regular"),
        )
    ])
    return res

def make_select_provider_callbacks():
    @callback(
        Input(kpsf("provider"), "value"),
        Output(ksfr("provider-active"), "data"),
        Output(kf("select-provider-config"), "data"),
        Output(kf("select-provider-config"), "value"),
        prevent_initial_callbacks=True,
    )
    def select_provider(providers):
        data = None
        active = []
        default = None

        if providers:
            active = providers
            data = make_select_data(providers)
            default = active[0]

        return active, data, default

def make_select():
    return dmc.Group([
        dmc.Switch(
            id=kf.sid("auto-apply"),
            offLabel=icons("autoff", width=20),
            onLabel=icons("auton", width=20),
            checked=True,
        ),
        dmc.Select(
            id=kgsf("provider"),
            placeholder="Select database",
            className="kmviz-gselect",
            size="xs"
        ),
        dmc.Select(
            id=kgsf("query"),
            placeholder="Select query",
            className="kmviz-gselect",
            size="xs"
        ),
    ], spacing=3, align="center", className="kmviz-gselect-group")

def make_select_callbacks():

    @callback(
        Input(kgsf("provider"), "value"),
        Output(kf.child("map").sid("select-preset"), "data"),
        Output(kf.child("plot").sid("select-preset"), "data"),
        Output(kf.child("map").sid("select-preset"), "style"),
        Output(kf.child("plot").sid("select-preset"), "style"),
    )
    def update_map_preset(value):
        if not value:
            return None, None, {"display": "none"}, {"display": "none"}

        if value.startswith("__kmviz_df"):
            prevent_update_on_none(None)

        p = state.kmstate.providers.get(value).presets

        map_data, map_style = None, {"display": "none"}
        plot_data, plot_style = None, {"display": "none"}
        if "map" in p:
            map_data = make_select_data(list(p["map"].keys()))
            map_style["display"] = "inline"
        if "plot" in p:
            plot_data = make_select_data(list(p["plot"].keys()))
            plot_style["display"] = "inline"

        return map_data, plot_data, map_style, plot_style




import uuid
import sys
from loguru import logger

import logging
logging.getLogger('werkzeug').setLevel(logging.ERROR)

def instance_id():
    return uuid.uuid4().hex

log_fmt = (
    "<red>[kmviz:" + instance_id() + "]</red> ~ "
    "<blue>{time:YYYY-MM-DD HH:mm:ss.SSS}</blue> ~ "
    "<level>{level: <4}</level> ~ "
    "<white><bold>{message}</bold></white>"
)

def setup_logger(log_level: str="DEBUG", log_directory: str=None, with_stderr: bool=True) -> None:
    log_level = log_level.upper()
    logger.remove()

    if with_stderr:
        logger.add(sys.stderr, format=log_fmt, colorize=True, level=log_level, backtrace=False, diagnose=False)

    if log_directory:
        logger.add(f"{log_directory}/kmviz-{instance_id()}-{{time:YYYY-MM-DD_HH:mm}}", rotation="1H", format=log_fmt)

def kmv_ex(exp) -> None:
    logger.exception(exp)

def kmv_debug(*args, **kwargs) -> None:
    logger.opt(depth=1).debug(*args, **kwargs)

def kmv_info(*args, **kwargs) -> None:
    logger.opt(depth=1).info(*args, **kwargs)

def kmv_warn(*args, **kwargs) -> None:
    logger.opt(depth=1).warning(*args, **kwargs)

def kmv_error(*args, **kwargs) -> None:
    logger.opt(depth=1).error(*args, **kwargs)




import dash_mantine_components as dmc
from dash_extensions.enrich import html

from kmviz.ui.layouts.index import make_index_layout, make_index_layout_callbacks, kindex
from kmviz.ui.layouts.table import make_table_layout, make_table_layout_callbacks, ktable
from kmviz.ui.layouts.map import make_map_layout, make_map_layout_callbacks, kmap
from kmviz.ui.layouts.sequence import make_sequence_layout, make_sequence_layout_callbacks, kseq
from kmviz.ui.layouts.plot import make_plot_layout, make_plot_layout_callbacks, kplot
from kmviz.ui.layouts.help import make_help_layout
from dash_iconify import DashIconify
from kmviz.ui import state
from kmviz.core.log import kmv_info

def _make_plugin_tab(plugin, name, icon):
    hide = {"display":"none"} if state.kmstate.plot_only else {}

    return dmc.Tab(
        name,
        value=f"{plugin}-{name}-value",
        id=f"kmviz-plugin-{plugin}-{name}-id",
        style=hide,
        icon=DashIconify(icon=icon) if icon else None
    )

def _make_plugin_panel(plugin, name, panel):
    return dmc.TabsPanel(
        panel, value=f"{plugin}-{name}-value"
    )

def make_tabs():
    hide = {"display":"none"} if state.kmstate.plot_only else {}

    pTabs = []
    pPanels = []

    for name, plugin in state.kmstate.plugins.items():
        layouts = plugin.layouts()
        if layouts:
            for tab_name, panel, icon in layouts:
                kmv_info(f"Load layout '{tab_name}' from plugin '{name}'")
                pTabs.append(_make_plugin_tab(name, tab_name, icon))
                pPanels.append(_make_plugin_panel(name, tab_name, panel))

    tabs = html.Div([
        dmc.Tabs([
            dmc.TabsList([
                dmc.Tab(
                    "Index",
                    value="index",
                    id=kindex.sid("panel"),
                    icon=DashIconify(icon="iconoir:db"),
                    style = hide),
                dmc.Tab(
                    "Table",
                    value="table",
                    disabled=True,
                    id=ktable.sid("panel"),
                    icon=DashIconify(icon="material-symbols:table")),
                dmc.Tab(
                    "Map",
                    value="map",
                    disabled=True,
                    id=kmap.sid("panel"),
                    icon=DashIconify(icon="fluent-mdl2:world"),),
                dmc.Tab(
                    "Plot",
                    value="plot",
                    disabled=True,
                    id=kplot.sid("panel"),
                    icon=DashIconify(icon="carbon:qq-plot"),),
                dmc.Tab(
                    "Sequence",
                    value="sequence",
                    disabled=True,
                    id=kseq.sid("panel"),
                    icon=DashIconify(icon="mdi:dna"),
                    style = hide),
                *pTabs,
                dmc.Tab(
                    "Help",
                    value="help",
                    icon=DashIconify(icon="material-symbols:help-outline")),
            ], className="kmviz-tab-header"),

            html.Div([

            dmc.TabsPanel(
                make_table_layout(), value="table"),
            dmc.TabsPanel(
                make_map_layout(), value="map"),
            dmc.TabsPanel(
                make_index_layout(), value="index"),
            dmc.TabsPanel(
                make_sequence_layout(), value="sequence"),
            dmc.TabsPanel(
                make_plot_layout(), value="plot"),
            *pPanels,
            dmc.TabsPanel(
                make_help_layout(), value="help"),
            ], className="kmviz-tabs")
        ], value="index" if not state.kmstate.plot_only else None, id="tab-select"),
    ])

    make_table_layout_callbacks()
    make_index_layout_callbacks()
    make_map_layout_callbacks()
    make_sequence_layout_callbacks()
    make_plot_layout_callbacks()

    return tabs




def make_server_cache_manager(params: dict):
    if params["type"] == "disk":
        from dash_extensions.enrich import FileSystemBackend
        return FileSystemBackend(**params["params"])
    elif params["type"] == "redis":
        from dash_extensions.enrich import RedisBackend
        return RedisBackend(**params["params"])

def make_result_cache_manager(params: dict):
    if params["type"] == "disk":
        from cachelib.file import FileSystemCache
        return FileSystemCache(**params["params"])
    elif params["type"] == "redis":
        from cachelib.redis import RedisCache
        return RedisCache(**params["params"])

def make_callback_cache_manager(params: dict):
    from dash import DiskcacheManager, CeleryManager
    if params["type"] == "disk":
        import diskcache
        return DiskcacheManager(diskcache.Cache(**params["params"]))
    elif params["type"] == "celery":
        from celery import Celery
        return CeleryManager(Celery(**params["params"]))



from dash import Patch, ALL, MATCH
from typing import Any, List, Tuple, Union

class IDFactory:
    def __init__(self, *prefixes) -> None:
        self._prefix = "-".join(prefixes)
        self._index = 0
        self._order = 0
        self._iid_register = set()
        self._sid_register = set()
        self._children = []

    def _iid(self, index: Any) -> dict:
        if index is None:
            res = { "type": self._prefix, "index": self._index}
            self._index += 1
        else:
            res = { "type": self._prefix, "index": index}

        self._iid_register.add((res["type"], res["index"], self._order))
        self._order += 1
        return res

    def sid(self, id: str) -> str:
        res = f"{self._prefix}-{id}"
        self._sid_register.add(res)
        return res

    def iid(self, index: Any=None) -> dict:
        return self._iid(index)

    def __call__(self, index: Any=None) -> dict:
        return self.iid(index)

    def child(self, suffix: str) -> "IDFactory":
        self._children.append(IDFactory(self._prefix, suffix))
        return self._children[-1]

    def list(self) -> Tuple[List[str], List[dict]]:
        return list(self._sid_register), list(self._iid_register)

    def list_all(self) -> Tuple[List[str], List[dict]]:
        sid, idd = self.list()
        for c in self._children:
            c_sid, c_iid = c.list_all()
            sid.extend(c_sid), idd.extend(c_iid)
        return sid, idd

    @property
    def type(self):
        return self._prefix

    @property
    def all(self):
        return {"type": self._prefix, "index": ALL}

kmviz_factory = IDFactory("kmviz")
kmviz_store_factory = kmviz_factory.child("store")



from .tsv import TsvMetaDB
from .sqlite import SQLiteMetaDB
from .mysql import MySQLMetaDB
from importlib import import_module
from kmviz.core import KmVizError

METADBS = {
    "tsv": TsvMetaDB,
    "sqlite": SQLiteMetaDB,
    "mysql": MySQLMetaDB,
}

def make_metadb_from_dict(d):
    if d["type"] not in METADBS:
        raise KmVizError(f"MetaDB type not recognized ('{d['type']}'), [{','.join(METADBS.keys())}]")

    return  METADBS[d["type"]](**d["params"])



